import java.awt.event.ActionListener;
import java.awt.event.ActionEvent;
import java.util.Random;


class Game implements ActionListener { // ActionListener is implemented so that the TTimer can trigger events here

	public Game(int level) {
		// Initializations
		d_control = new InputHandler(this); // Handles input and informs us

		d_blocks = new Block[d_width * d_height];
		for (int x = 0; x < d_width; ++x)
			for (int y = 0; y < d_height; ++y) 
				// Initting the blocks
				d_blocks[y*d_width + x] = new Block();

		// Adding random blocks on the bottom based on level. (3 per each level)
		for (int blocks_left = 3*level; blocks_left > 0; --blocks_left) {
			int pile = d_rndgen.nextInt(d_width); // Which pile to add on

			// We're adding the block to the next available slot
			int freeslot = 0;
			while (d_blocks[freeslot*d_width + pile].getColor() != Block.Color.EMPTY) freeslot++;

			// If we were trying to add to a pile that is more than 10 blocks high already, we rerandomize
			if (freeslot > 9) {
				blocks_left++;
				continue;
			}

			// Adding the block
			d_blocks[freeslot*d_width + pile] = Block.getRand();
		}


		d_gameboard = new Visual("Gameloft Tetris Wannabe", d_blocks, d_width, d_height);
		d_gameboard.addKeyListener(d_control);	// The Visual is the window that has focus and therefore listens to keyboard

		// A little randomness is added by evaluating the randomly created blocks --
		// part of them might get deleted in the beginning :-)
		evaluateBoard();

		d_blocktimer = new TTimer((10-level)*1000, this); // One second delay for lvl10, 11 second delay for lvl1
		d_droptimer = new TTimer(500, this); // The delay gets set later on
		d_2mintimer = new TTimer(2*60*1000, this);
		// An automatic run might be in place
		run();
	}

	// So far this catches TTimer events.  Acts as a hub.
	public void actionPerformed(ActionEvent ae) {
		if (ae.getSource() == d_droptimer)
			dropBlocks();
		if (ae.getSource() == d_blocktimer)
			spawnBlocks();

		// If the 2min timer ticks, we reduce spawn time
		if (ae.getSource() == d_2mintimer)
			d_blocktimer.setDelay((int)(d_blocktimer.getDelay() * d_lvl_spawn_delay));
	}
	public void keyEvent(String keyname) {
		switch (d_mode) {
			case 0: // We're moving the cursor within the board
				if (keyname.equals("Up") &&
						d_cursor/d_width < d_height-1)
					d_gameboard.setCursor(d_cursor += d_width);

				if (keyname.equals("Down") &&
						d_cursor/d_width > 0)
					d_gameboard.setCursor(d_cursor -= d_width);

				if (keyname.equals("Right") &&
						d_cursor%d_width < d_width-1)
					d_gameboard.setCursor(++d_cursor);

				if (keyname.equals("Left") &&
						d_cursor%d_width > 0)
					d_gameboard.setCursor(--d_cursor);

				if (keyname.equals("Space")) // We selected a block
					// If it isn't empty or moving, we change modes
					if (d_blocks[d_cursor].getColor() != Block.Color.EMPTY &&
							d_blocks[d_cursor].isStatic()) 
						d_mode = 1;

				break;

			case 1:
				// We're selecting a direction to swap a block.  We check that we're not hitting borders,
				// and swapBlocks() is doing the rest.  If it returns true, the swap was successful.
				if (keyname.equals("Up") &&
						d_cursor/d_width < d_height-1)
					if (swapBlocks(d_cursor, d_cursor+d_width))
						d_mode = 0; // Back to the normal mode	

				if (keyname.equals("Down") &&
						d_cursor/d_width > 0)
					if (swapBlocks(d_cursor, d_cursor-d_width))
						d_mode = 0; // Back to the normal mode	

				if (keyname.equals("Right") &&
						d_cursor%d_width < d_width-1)
					if (swapBlocks(d_cursor, d_cursor+1))
						d_mode = 0; // Back to the normal mode	

				if (keyname.equals("Left") &&
						d_cursor%d_width > 0)
					if (swapBlocks(d_cursor, d_cursor-1))
						d_mode = 0; // Back to the normal mode	

				if (keyname.equals("Space")) // Aborting this mode
					d_mode = 0;
	
				break;
		}
	}

	// False if the swap isn't possible
	public boolean swapBlocks(int i1, int i2) {
		if (d_blocks[i2].getColor() == Block.Color.EMPTY ||
				!d_blocks[i2].isStatic()) // If it is empty or moving, we can't do this
			return false;

		// Swapping the blocks
		Block.Color temp = d_blocks[i1].getColor();
		d_blocks[i1].setColor(d_blocks[i2].getColor());
		d_blocks[i2].setColor(temp);
	
		// Evaluation and refresh	
		evaluateBoard();
		d_gameboard.refresh();

		return true;
	}
	
	public void run() {
		// We start the timers and drop our first blocks
		d_droptimer.start();
		d_blocktimer.start();
		d_2mintimer.start();
		spawnBlocks();
	}

	// A new block pair is spawned here.  This is the only place where the game can end.
	public void spawnBlocks() {
		if (d_game_over)
			return;

		// One point for each block pair spawned.  This equals the game played but favours higher levels.

		// The block pair can start from 0 to d_width-2.  d_width-1 is thus the proper argument for nextInt
		int startpos = d_rndgen.nextInt(d_width-1);
		
		// The game is over if there's a block in either (startpos, d_height-1) or (startpos+1, d_height-1)
		int index1 = (d_height-1)*d_width + startpos;
		int index2 = (d_height-1)*d_width + startpos + 1;
		if (d_blocks[index1].getColor() != Block.Color.EMPTY || d_blocks[index2].getColor() != Block.Color.EMPTY) {
			d_game_over = true;
			d_gameboard.reportScore(d_score);
		}

		// Spawning
		d_blocks[index1] = Block.getRand();
		d_blocks[index2] = Block.getRand();
		d_blocks[index1].setStatic(false); // They're not static in the beginning
		d_blocks[index2].setStatic(false);

		// Droptimer gets reset as well
		d_droptimer.setDelay(d_blocktimer.getDelay() / d_height); // It drops at the same rate no matter where it lands

		d_gameboard.refresh(); // Refreshing the visual
	}

	// This method is used for dropping all non-static blocks by one
	public void dropBlocks() {
		for (int x = 0; x < d_width; ++x)
			for (int y = 0; y < d_height; ++y)
				if (!d_blocks[y*d_width + x].isStatic()) { // This one is dropping

					// If there's a block or border beneath this object, it turns static
					if (y == 0 || // If y is 0, this matches, and the following condition never gets evaluated, so y-1 wont be a problem
							!d_blocks[(y-1)*d_width + x].getColor().equals(Block.Color.EMPTY)) { // This is never not-static so this'll be okay
						d_blocks[y*d_width + x].setStatic(true);

						// After it has turned static, an evaluation of the board is in place
						evaluateBoard();

					} else { // It drops by one
						d_blocks[(y-1)*d_width + x].setColor(d_blocks[y*d_width + x].getColor());
						d_blocks[(y-1)*d_width + x].setStatic(false); // Remains non-static
						d_blocks[y*d_width + x].setColor(Block.Color.EMPTY);
						d_blocks[y*d_width + x].setStatic(true);

						d_gameboard.refresh(); // We need to refresh the visual
					}
				}
	}

	// This one checks if there's acceptable shapes, scores them, and drops the necessary blocks.
	public void evaluateBoard() {
		// Blocks can still be moved around, if the game has ended, but nothing gets evaluated or dropped
		if (d_game_over) 
			return;

		boolean deleted = false;

		// Going through the blocks one by one
		for (int x = 0; x < d_width; ++x)
			for (int y = 0; y < d_height; ++y) 
				if (TetrisBlock.match(d_blocks, d_width, d_height, x, y))
					// Now there's a match
					deleted = true;

		// If blocks were deleted, we need to drop all static blocks and re-evaluate
		if (deleted) {
			// Actually deleting the blocks (we postponed the deletion until now so overlaps are handled correctly)
			for (int x = 0; x < d_width; ++x)
				for (int y = 0; y < d_height; ++y) {
					d_blocks[y*d_width + x].clear(); // Clears if marked for deletion
					// One point for each cleared block
					d_score++;
				}

			// Dropping
			for (int x = 0; x < d_width; ++x)
				for (int y = 1; y < d_height; ++y) // First row can't drop (y = 0)
					if (d_blocks[y*d_width + x].isStatic() && // Needs to be static
						d_blocks[y*d_width + x].getColor() != Block.Color.EMPTY) { // And something but empty

						// Now we're looping downwards until we hit floor or another block
						int down;
						for (down = y-1; 
								down >= 0 && d_blocks[down*d_width + x].getColor() == Block.Color.EMPTY;
								--down);
						// We're not executing anything in the loop, just dropping down
						if (down != y-1) { // If there was any room to begin with
							d_blocks[(down+1)*d_width + x].setColor(
									d_blocks[y*d_width + x].getColor());
							d_blocks[y*d_width + x].setColor(Block.Color.EMPTY);
						}
					}


			d_gameboard.refresh();
			evaluateBoard();
		}
	}


	/* Data */
	private TTimer d_droptimer, d_blocktimer, d_2mintimer; // Droptimer ticks for each block drop, and blocktimer ticks for a new block spawn
	private InputHandler d_control;
	private Visual d_gameboard;

	private final int d_width = 6, d_height = 12; // The board size
	private Block[] d_blocks; // Shared with Visual
	private int d_cursor = (d_height/2)*d_width + d_width/2; // About in the middle
	private int d_mode = 0; // 0: Cursor can be moved,  1: Block is selected, select swap direction

	private boolean d_game_over = false;
	private Random d_rndgen = new Random(); // It is seeded with current time by default, which is sufficient.

	private final float d_lvl_spawn_delay = 0.8f; // How much the spawn delay drops every 2mins (0.8 == by 20%)
	private int d_score = 0;
}
