// Enums are efficient in Java and suit our purposes here
public enum TetrisBlock {
	// We could only describe the main blocks, and then try to match rotations automatically.
	// This would not, however, be as scalable.

	// This is the 2x2 block
	A (2, 2, new boolean[] {
		true, 	true, 
		true, 	true}),
	// 1x4 block, 4 blocks on top of each others.  
	B (1, 4, new boolean[] {
		true,
		true, 
		true,
		true}),
	// The same as above, but sideways
	C (4, 1, new boolean[] {
		true, true, true, true }),
 	// L upside down.  You probably got the idea  	
	D (2, 3, new boolean[] {
		true,	true,
		true,	false,
		true,	false}),
	E (2, 3, new boolean[] {
		true,	false,
		true,	false,
		true,	true}),
	F (2, 3, new boolean[] {
		false,	true,
		false,	true,
		true,	true}),
	G (2, 3, new boolean[] {
		true,	true,
		false,	true,
		false,	true}),
	H (3, 2, new boolean[] {
		true,	true,	true,
		false,	false, 	true}),
	I (3, 2, new boolean[] {
		true,	true,	true,
		true,	false, 	false}),
	J (3, 2, new boolean[] {
		true,	false,	false,
		true,	true, 	true}),
	K (3, 2, new boolean[] {
		false,	false,	true,
		true,	true, 	true}),
	L (3, 2, new boolean[] {
		false,	true,	true,
		true,	true, 	false}),
	M (3, 2, new boolean[] {
		true,	true,	false,
		false,	true, 	true}),
	N (2, 3, new boolean[] {
		true,	false,
		true,	true,
		false,	true}),
	O (2, 3, new boolean[] {
		false,	true,
		true,	true,
		true,	false});


	// W and h tell the dimensions of the block, while blocks tells the occupied spots
	private int d_w, d_h;
	private boolean[] d_blocks;
	private TetrisBlock(int w, int h, boolean[] blocks) {
		d_w = w;
		d_h = h;
		d_blocks = blocks;
	}

	// This function checks whether (any) TetrisBlock is found from blocklist at the given position
	// All matches are marked to be deleted 
	public static boolean match(Block[] blocklist, int w, int h, int x, int y) { // w,h refers to the blocklist dimensions
		boolean found = false;

		for (TetrisBlock tb : TetrisBlock.values()) { // Going through all blocks
			Block.Color color = Block.Color.EMPTY; // We start with empty

			// If there's no room for the tetris block, we skip it as it can't match
			if (x+tb.d_w > w ||
				y+tb.d_h > h)
				continue;

			boolean blockfail = false;
			// We loop through the blocks in our tetris block and check for similarities in blocklist	
			for (int x_i = x; x_i < x + tb.d_w; ++x_i) 
				for (int y_i = y; y_i < y + tb.d_h && !blockfail; ++y_i) { // We bail out in this loop if the match fails

					int block_x = x_i - x; // Coordinates for the tetrisblock
					int block_y = y_i - y;

					// Let's see if there's an occupied block
					if (tb.d_blocks[block_y*tb.d_w + block_x]) {
						// We fail if color is set (this isn't the first block)
						// and it doesn't match
						if (color != Block.Color.EMPTY &&
							blocklist[y_i*w + x_i].getColor() != color) {
							blockfail = true;
						} else {
							// We also fail if the block is empty (this should be occupied) or it's moving
							if (blocklist[y_i*w + x_i].getColor() == Block.Color.EMPTY ||
									!blocklist[y_i*w + x_i].isStatic()) {
								blockfail = true;
							} else {
								// This block is acceptable
								// Setting the new color (it does match, or we wouldn't be here)
								color = blocklist[y_i*w + x_i].getColor();
							}
						}
					}
				}

			if (!blockfail) { // We found this block, so we're erasing it
				found = true; // At least one found

				for (int x_i = x; x_i < x + tb.d_w; ++x_i) 
					for (int y_i = y; y_i < y + tb.d_h; ++y_i) 
						if (tb.d_blocks[(y_i-y)*tb.d_w + x_i-x]) // If it is occupied
							blocklist[y_i*w + x_i].setDeleted(true); // We mark it deleted
			}
				
		}

		return found;
	}
}
